@@ -9,11 +9,12 @@ require Exporter;
 our @ISA = qw(Exporter);
 
 our $VERBOSE = 0;        # verbose mode
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 # Arithmetic Coding settings
-use constant BITS => 32;
-use constant MAX  => oct('0b' . ('1' x BITS));
+use constant BITS         => 32;
+use constant MAX          => oct('0b' . ('1' x BITS));
+use constant INITIAL_FREQ => 1;
 
 our %EXPORT_TAGS = (
     'all' => [
@@ -49,8 +50,13 @@ our %EXPORT_TAGS = (
           mtf_encode
           mtf_decode
 
-          mtf_encode_alphabet
-          mtf_decode_alphabet
+          encode_alphabet
+          decode_alphabet
+
+          run_length
+
+          binary_vrl_encode
+          binary_vrl_decode
 
           rle4_encode
           rle4_decode
@@ -76,19 +82,27 @@ our %EXPORT_TAGS = (
           lz77_encode
           lz77_decode
 
-          _create_cfreq
           ac_encode
           ac_decode
 
           create_ac_entry
           decode_ac_entry
 
+          adaptive_ac_encode
+          adaptive_ac_decode
+
+          create_adaptive_ac_entry
+          decode_adaptive_ac_entry
+
           abc_encode
           abc_decode
 
           fibonacci_encode
           fibonacci_decode
 
+          elias_gamma_encode
+          elias_gamma_decode
+
           elias_omega_encode
           elias_omega_decode
 
@@ -284,6 +298,47 @@ sub fibonacci_decode ($fh) {
     return \@symbols;
 }
 
+#####################
+# Elias gamma coding
+#####################
+
+sub elias_gamma_encode ($integers) {
+
+    my $bitstring = '';
+    foreach my $k (scalar(@$integers), @$integers) {
+        my $t = sprintf('%b', $k + 1);
+        $bitstring .= ('1' x (length($t) - 1)) . '0' . substr($t, 1);
+    }
+
+    pack('B*', $bitstring);
+}
+
+sub elias_gamma_decode ($fh) {
+
+    if (ref($fh) eq '') {
+        open my $fh2, '<:raw', \$fh;
+        return __SUB__->($fh2);
+    }
+
+    my @ints;
+    my $len    = 0;
+    my $buffer = '';
+
+    for (my $k = 0 ; $k <= $len ; ++$k) {
+
+        my $bl = 0;
+        ++$bl while (read_bit($fh, \$buffer) eq '1');
+
+        push @ints, oct('0b' . '1' . join('', map { read_bit($fh, \$buffer) } 1 .. $bl)) - 1;
+
+        if ($k == 0) {
+            $len = pop(@ints);
+        }
+    }
+
+    return \@ints;
+}
+
 #####################
 # Elias omega coding
 #####################
@@ -539,11 +594,11 @@ sub _create_cfreq ($freq) {
     return (\@cf, $T);
 }
 
-sub ac_encode ($bytes_arr) {
+sub ac_encode ($symbols) {
 
     my $enc        = '';
-    my $EOF_SYMBOL = (max(@$bytes_arr) // 0) + 1;
-    my @bytes      = (@$bytes_arr, $EOF_SYMBOL);
+    my $EOF_SYMBOL = (max(@$symbols) // 0) + 1;
+    my @bytes      = (@$symbols, $EOF_SYMBOL);
 
     my %freq;
     ++$freq{$_} for @bytes;
@@ -679,9 +734,9 @@ sub ac_decode ($fh, $freq) {
     return \@dec;
 }
 
-sub create_ac_entry ($bytes, $out_fh = undef) {
+sub create_ac_entry ($symbols, $out_fh = undef) {
 
-    my ($enc, $freq) = ac_encode($bytes);
+    my ($enc, $freq) = ac_encode($symbols);
     my $max_symbol = max(keys %$freq) // 0;
 
     my @freqs;
@@ -699,6 +754,11 @@ sub create_ac_entry ($bytes, $out_fh = undef) {
 
 sub decode_ac_entry ($fh) {
 
+    if (ref($fh) eq '') {
+        open my $fh2, '<:raw', \$fh;
+        return __SUB__->($fh2);
+    }
+
     my @freqs    = @{delta_decode($fh)};
     my $bits_len = pop(@freqs);
 
@@ -720,6 +780,210 @@ sub decode_ac_entry ($fh) {
     return [];
 }
 
+#############################################
+# Adaptive Arithemtic Coding (in fixed bits)
+#############################################
+
+sub _create_adaptive_cfreq ($freq_value, $alphabet) {
+
+    my $T = 0;
+    my (@cf, @freq);
+
+    foreach my $i (@$alphabet) {
+        $freq[$i] = $freq_value;
+        $cf[$i]   = $T;
+        $T += $freq_value;
+        $cf[$i + 1] = $T;
+    }
+
+    return (\@freq, \@cf, $T);
+}
+
+sub _increment_freq ($c, $alphabet, $freq, $cf) {
+
+    ++$freq->[$c];
+    my $T = $cf->[$c];
+
+    foreach my $i (@$alphabet) {
+        next if ($i < $c);
+        $cf->[$i] = $T;
+        $T += $freq->[$i];
+        $cf->[$i + 1] = $T;
+    }
+
+    return $T;
+}
+
+sub adaptive_ac_encode ($symbols) {
+
+    my $enc      = '';
+    my @bytes    = (@$symbols, (max(@$symbols) // 0) + 1);
+    my $alphabet = [sort { $a <=> $b } uniq(@bytes)];
+
+    my ($freq, $cf, $T) = _create_adaptive_cfreq(INITIAL_FREQ, $alphabet);
+
+    if ($T > MAX) {
+        die "Too few bits: $T > ${\MAX}";
+    }
+
+    my $low      = 0;
+    my $high     = MAX;
+    my $uf_count = 0;
+
+    foreach my $c (@bytes) {
+
+        my $w = $high - $low + 1;
+
+        $high = ($low + int(($w * $cf->[$c + 1]) / $T) - 1) & MAX;
+        $low  = ($low + int(($w * $cf->[$c]) / $T)) & MAX;
+
+        $T = _increment_freq($c, $alphabet, $freq, $cf);
+
+        if ($high > MAX) {
+            die "high > MAX: $high > ${\MAX}";
+        }
+
+        if ($low >= $high) { die "$low >= $high" }
+
+        while (1) {
+
+            if (($high >> (BITS - 1)) == ($low >> (BITS - 1))) {
+
+                my $bit = $high >> (BITS - 1);
+                $enc .= $bit;
+
+                if ($uf_count > 0) {
+                    $enc .= join('', 1 - $bit) x $uf_count;
+                    $uf_count = 0;
+                }
+
+                $low <<= 1;
+                ($high <<= 1) |= 1;
+            }
+            elsif (((($low >> (BITS - 2)) & 0x1) == 1) && ((($high >> (BITS - 2)) & 0x1) == 0)) {
+                ($high <<= 1) |= (1 << (BITS - 1));
+                $high |= 1;
+                ($low <<= 1) &= ((1 << (BITS - 1)) - 1);
+                ++$uf_count;
+            }
+            else {
+                last;
+            }
+
+            $low  &= MAX;
+            $high &= MAX;
+        }
+    }
+
+    $enc .= '0';
+    $enc .= '1';
+
+    while (length($enc) % 8 != 0) {
+        $enc .= '1';
+    }
+
+    return ($enc, $alphabet);
+}
+
+sub adaptive_ac_decode ($fh, $alphabet) {
+
+    if (ref($fh) eq '') {
+        open my $fh2, '<:raw', \$fh;
+        return __SUB__->($fh2, $alphabet);
+    }
+
+    my ($freq, $cf, $T) = _create_adaptive_cfreq(INITIAL_FREQ, $alphabet);
+
+    my @dec;
+    my $low        = 0;
+    my $high       = MAX;
+    my @alpha      = @$alphabet;
+    my $max_symbol = $alpha[-1];
+
+    my $enc = oct('0b' . join '', map { getc($fh) // 1 } 1 .. BITS);
+
+    while (1) {
+        my $w  = ($high + 1) - $low;
+        my $ss = int((($T * ($enc - $low + 1)) - 1) / $w);
+
+        my $i = 0;
+        foreach my $j (@alpha) {
+            if ($cf->[$j] <= $ss and $ss < $cf->[$j + 1]) {
+                $i = $j;
+                last;
+            }
+        }
+
+        last if ($i == $max_symbol);
+        push @dec, $i;
+
+        $high = ($low + int(($w * $cf->[$i + 1]) / $T) - 1) & MAX;
+        $low  = ($low + int(($w * $cf->[$i]) / $T)) & MAX;
+
+        $T = _increment_freq($i, $alphabet, $freq, $cf);
+
+        if ($high > MAX) {
+            die "high > MAX: ($high > ${\MAX})";
+        }
+
+        if ($low >= $high) { die "$low >= $high" }
+
+        while (1) {
+
+            if (($high >> (BITS - 1)) == ($low >> (BITS - 1))) {
+                ($high <<= 1) |= 1;
+                $low <<= 1;
+                ($enc <<= 1) |= (getc($fh) // 1);
+            }
+            elsif (((($low >> (BITS - 2)) & 0x1) == 1) && ((($high >> (BITS - 2)) & 0x1) == 0)) {
+                ($high <<= 1) |= (1 << (BITS - 1));
+                $high |= 1;
+                ($low <<= 1) &= ((1 << (BITS - 1)) - 1);
+                $enc = (($enc >> (BITS - 1)) << (BITS - 1)) | (($enc & ((1 << (BITS - 2)) - 1)) << 1) | (getc($fh) // 1);
+            }
+            else {
+                last;
+            }
+
+            $low  &= MAX;
+            $high &= MAX;
+            $enc  &= MAX;
+        }
+    }
+
+    return \@dec;
+}
+
+sub create_adaptive_ac_entry ($symbols, $out_fh = undef) {
+
+    my ($enc, $alphabet) = adaptive_ac_encode($symbols);
+
+    $out_fh // open $out_fh, '>:raw', \my $out_str;
+    print $out_fh pack('N', length($enc));
+    print $out_fh encode_alphabet($alphabet);
+    print $out_fh pack("B*", $enc);
+    return $out_str;
+}
+
+sub decode_adaptive_ac_entry ($fh) {
+
+    if (ref($fh) eq '') {
+        open my $fh2, '<:raw', \$fh;
+        return __SUB__->($fh2);
+    }
+
+    my $enc_len  = unpack('N', join('', map { getc($fh) // die "error" } 1 .. 4));
+    my $alphabet = decode_alphabet($fh);
+
+    if ($enc_len > 0) {
+        my $bits = read_bits($fh, $enc_len);
+        open my $bits_fh, '<:raw', \$bits;
+        return adaptive_ac_decode($bits_fh, $alphabet);
+    }
+
+    return [];
+}
+
 ##########################
 # Move to front transform
 ##########################
@@ -817,6 +1081,82 @@ sub bwt_decode ($bwt, $idx) {    # fast inversion
     return $dec;
 }
 
+#####################
+# Generic run-length
+#####################
+
+sub run_length ($arr, $max_run = undef) {
+
+    @$arr || return [];
+
+    my @result     = [$arr->[0], 1];
+    my $prev_value = $arr->[0];
+
+    foreach my $i (1 .. $#$arr) {
+
+        my $curr_value = $arr->[$i];
+
+        if ($curr_value eq $prev_value and (defined($max_run) ? $result[-1][1] < $max_run : 1)) {
+            ++$result[-1][1];
+        }
+        else {
+            push(@result, [$curr_value, 1]);
+        }
+
+        $prev_value = $curr_value;
+    }
+
+    return \@result;
+}
+
+######################################
+# Binary variable run-length encoding
+######################################
+
+sub binary_vrl_encode ($bitstring) {
+
+    my @bits    = split(//, $bitstring);
+    my $encoded = $bits[0];
+
+    foreach my $rle (@{run_length(\@bits)}) {
+        my ($c, $v) = @$rle;
+
+        if ($v == 1) {
+            $encoded .= '0';
+        }
+        else {
+            my $t = sprintf('%b', $v - 1);
+            $encoded .= join('', '1' x length($t), '0', substr($t, 1));
+        }
+    }
+
+    return $encoded;
+}
+
+sub binary_vrl_decode ($bitstring) {
+
+    my $decoded = '';
+    my $bit     = substr($bitstring, 0, 1, '');
+
+    while ($bitstring ne '') {
+
+        $decoded .= $bit;
+
+        my $bl = 0;
+        while (substr($bitstring, 0, 1, '') eq '1') {
+            ++$bl;
+        }
+
+        if ($bl > 0) {
+            $decoded .= $bit x oct('0b1' . join('', map { substr($bitstring, 0, 1, '') } 1 .. $bl - 1));
+        }
+
+        $bit = ($bit eq '1' ? '0' : '1');
+    }
+
+    return $decoded;
+}
+
 #####################
 # RLE4 used in Bzip2
 #####################
@@ -976,7 +1316,7 @@ sub _mtf_encode_alphabet_256 ($alphabet) {
         }
     }
 
-    my $delta = delta_encode([@marked], 1);
+    my $delta = delta_encode(\@marked, 1);
 
     $VERBOSE && say STDERR "Populated : ", sprintf('%08b', $populated);
     $VERBOSE && say STDERR "Marked    : @marked";
@@ -1078,7 +1418,7 @@ sub bwt_decode_symbolic ($bwt, $idx) {    # fast inversion
     return \@dec;
 }
 
-sub mtf_encode_alphabet ($alphabet) {
+sub encode_alphabet ($alphabet) {
 
     my $max_symbol = max(@$alphabet);
 
@@ -1090,7 +1430,12 @@ sub mtf_encode_alphabet ($alphabet) {
     return (chr(0) . delta_encode([reverse @$alphabet]));
 }
 
-sub mtf_decode_alphabet ($fh) {
+sub decode_alphabet ($fh) {
+
+    if (ref($fh) eq '') {
+        open my $fh2, '<:raw', \$fh;
+        return __SUB__->($fh2);
+    }
 
     if (ord(getc($fh) // die "error") == 1) {
         return _mtf_decode_alphabet_256($fh);
@@ -1114,12 +1459,12 @@ sub bz2_compress_symbolic ($symbols, $out_fh = undef, $entropy_sub = \&create_hu
 
     my @bytes        = @$bwt;
     my @alphabet     = sort { $a <=> $b } uniq(@bytes);
-    my $alphabet_enc = mtf_encode_alphabet(\@alphabet);
+    my $alphabet_enc = encode_alphabet(\@alphabet);
 
     $VERBOSE && say STDERR "BWT index = $idx";
     $VERBOSE && say STDERR "Max symbol: ", max(@alphabet) // 0;
 
-    my $mtf = mtf_encode(\@bytes, [@alphabet]);
+    my $mtf = mtf_encode(\@bytes, \@alphabet);
     my $rle = zrle_encode($mtf);
 
     $out_fh // open $out_fh, '>:raw', \my $out_str;
@@ -1139,7 +1484,7 @@ sub bz2_decompress_symbolic ($fh, $entropy_sub = \&decode_huffman_entry) {
     }
 
     my $idx      = unpack('N', join('', map { getc($fh) // die "error" } 1 .. 4));
-    my $alphabet = mtf_decode_alphabet($fh);
+    my $alphabet = decode_alphabet($fh);
 
     $VERBOSE && say STDERR "BWT index = $idx";
     $VERBOSE && say STDERR "Alphabet size: ", scalar(@$alphabet);
@@ -1162,9 +1507,9 @@ sub bz2_compress ($chunk, $out_fh = undef, $entropy_sub = \&create_huffman_entry
 
     my @bytes        = unpack('C*', $bwt);
     my @alphabet     = sort { $a <=> $b } uniq(@bytes);
-    my $alphabet_enc = mtf_encode_alphabet(\@alphabet);
+    my $alphabet_enc = encode_alphabet(\@alphabet);
 
-    my $mtf = mtf_encode(\@bytes, [@alphabet]);
+    my $mtf = mtf_encode(\@bytes, \@alphabet);
     my $rle = zrle_encode($mtf);
 
     $out_fh // open $out_fh, '>:raw', \my $out_str;
@@ -1184,7 +1529,7 @@ sub bz2_decompress ($fh, $out_fh = undef, $entropy_sub = \&decode_huffman_entry)
     }
 
     my $idx      = unpack('N', join('', map { getc($fh) // return undef } 1 .. 4));
-    my $alphabet = mtf_decode_alphabet($fh);
+    my $alphabet = decode_alphabet($fh);
 
     $VERBOSE && say STDERR "BWT index = $idx";
     $VERBOSE && say STDERR "Alphabet size: ", scalar(@$alphabet);
@@ -1263,7 +1608,7 @@ sub lzss_encode ($str) {
     my @chars  = split(//, $str);
     my $end    = $#chars;
 
-    my $min_len = 3;                          # $LENGTH_SYMBOLS->[0][0];
+    my $min_len = 4;                          # $LENGTH_SYMBOLS->[0][0];
     my $max_len = $LENGTH_SYMBOLS->[-1][0];
 
     my %literal_freq;
@@ -1669,7 +2014,7 @@ This functionality is provided by the function C<bz2_compress()>, which can be e
 
     my @bytes        = unpack('C*', $bwt);
     my @alphabet     = sort { $a <=> $b } uniq(@bytes);
-    my $alphabet_enc = mtf_encode_alphabet(\@alphabet);
+    my $alphabet_enc = encode_alphabet(\@alphabet);
 
     my $mtf = mtf_encode(\@bytes, \@alphabet);
     my $rle = zrle_encode($mtf);
@@ -1710,9 +2055,9 @@ An array of symbols means an array of non-negative integer values.
 
 =head3 filehandle
 
-An input filehandle is denoted by C<$fh>, while an output filehandle is denoted by C<$out_fh>.
+An input filehandle is denoted by C<$fh>, while an output file-handle is denoted by C<$out_fh>.
 
-The encoding of input and output filehandles must be set to C<:raw>.
+The encoding of input and output file-handles must be set to C<:raw>.
 
 =head1 HIGH-LEVEL FUNCTIONS
 
@@ -1722,6 +2067,9 @@ The encoding of input and output filehandles must be set to C<:raw>.
       create_ac_entry(\@symbols, $fh)      # Create an Arithmetic Coding block
       decode_ac_entry($fh)                 # Decode an Arithmetic Coding block
 
+      create_adaptive_ac_entry(\@symbols, $fh)  # Create an Adaptive Arithmetic Coding block
+      decode_adaptive_ac_entry($fh)             # Decode an Adaptive Arithmetic Coding block
+
       bz2_compress($string)                # Bzip2-like compression (RLE4+BWT+MTF+ZRLE+Huffman coding)
       bz2_decompress($fh)                  # Inverse of the above method
 
@@ -1742,13 +2090,16 @@ The encoding of input and output filehandles must be set to C<:raw>.
       delta_encode(\@ints, $double=0)      # Delta encoding of an array of ints
       delta_decode($fh, $double=0)         # Inverse of the above method
 
-      fibonacci_encode(\@ints)             # Fibonacci coding of an array of non-negative ints
+      fibonacci_encode(\@symbols)          # Fibonacci coding of an array of symbols
       fibonacci_decode($fh)                # Inverse of the above method
 
-      elias_omega_encode(\@ints)           # Elias Omega coding method of an array non-negative ints
+      elias_gamma_encode(\@symbols)        # Elias Gamma coding method of an array of symbols
+      elias_gamma_decode($fh)              # Inverse of the above method
+
+      elias_omega_encode(\@symbols)        # Elias Omega coding method of an array of symbols
       elias_omega_decode($fh)              # Inverse of the above method
 
-      abc_encode(\@ints)                   # Adaptive Binary Concatenation method of an array of non-negative ints
+      abc_encode(\@symbols)                # Adaptive Binary Concatenation method of an array of symbols
       abc_decode($fh)                      # Inverse of the above method
 
       bwt_encode($string)                  # Burrows-Wheeler transform
@@ -1760,8 +2111,10 @@ The encoding of input and output filehandles must be set to C<:raw>.
       mtf_encode(\@symbols, \@alphabet)    # Move-to-front transform
       mtf_decode(\@mtf, \@alphabet)        # Inverse of the above method
 
-      mtf_encode_alphabet(\@alphabet)      # Encode the Move-to-front alphabet
-      mtf_decode_alphabet($fh)             # Decode the Move-to-front alphabet
+      encode_alphabet(\@alphabet)          # Encode an alphabet of symbols into a binary string
+      decode_alphabet($fh)                 # Inverse of the above method
+
+      run_length(\@symbols, $max=undef)    # Run-length encoding, returning a 2D array
 
       rle4_encode(\@symbols, $max=255)     # Run-length encoding with 4 or more consecutive characters
       rle4_decode(\@rle4)                  # Inverse of the above method
@@ -1772,13 +2125,19 @@ The encoding of input and output filehandles must be set to C<:raw>.
       ac_encode(\@symbols)                 # Arithmetic Coding applied on an array of symbols
       ac_decode($bitstring, \%freq)        # Inverse of the above method
 
+      adaptive_ac_encode(\@symbols)               # Adaptive Arithmetic Coding applied on an array of symbols
+      adaptive_ac_decode($bitstring, \@alphabet)  # Inverse of the above method
+
       lzw_encode($string)                  # LZW encoding of a given string
       lzw_decode(\@symbols)                # Inverse of the above method
 
 =head1 LOW-LEVEL FUNCTIONS
 
-      read_bit($fh, \$buffer)              # Read one bit from filehandle
-      read_bits($fh, $len)                 # Read $len bits from filehandle
+      read_bit($fh, \$buffer)              # Read one bit from file-handle
+      read_bits($fh, $len)                 # Read `$len` bits from file-handle
+
+      binary_vrl_encode($bitstring)        # Binary variable run-length encoding
+      binary_vrl_decode($bitstring)        # Binary variable run-length decoding
 
       bwt_sort($string)                    # Burrows-Wheeler sorting
       bwt_sort_symbolic(\@symbols)         # Burrows-Wheeler sorting, applied on an array of symbols
@@ -1806,9 +2165,9 @@ The encoding of input and output filehandles must be set to C<:raw>.
 
 High-level function that generates a Huffman coding block.
 
-It takes two parameters: C<\@symbols>, which represents the symbols to be encoded, and C<$out_fh>, which is optional, and represents the filehandle where to write the result.
+It takes two parameters: C<\@symbols>, which represents the symbols to be encoded, and C<$out_fh>, which is optional, and represents the file-handle where to write the result.
 
-When the second parameter is ommited, the function returns a binary string.
+When the second parameter is omitted, the function returns a binary string.
 
 =head2 decode_huffman_entry
 
@@ -1824,9 +2183,9 @@ Inverse of C<create_huffman_entry()>.
 
 High-level function that generates an Arithmetic Coding block.
 
-It takes two parameters: C<\@symbols>, which represents the symbols to be encoded, and C<$out_fh>, which is optional, and represents the filehandle where to write the result.
+It takes two parameters: C<\@symbols>, which represents the symbols to be encoded, and C<$out_fh>, which is optional, and represents the file-handle where to write the result.
 
-When the second parameter is ommited, the function returns a binary string.
+When the second parameter is omitted, the function returns a binary string.
 
 =head2 decode_ac_entry
 
@@ -1835,9 +2194,27 @@ When the second parameter is ommited, the function returns a binary string.
 
 Inverse of C<create_ac_entry()>.
 
+=head2 create_adaptive_ac_entry
+
+    create_adaptive_ac_entry(\@symbols, $out_fh);        # writes to $out_fh
+    my $string = create_adaptive_ac_entry(\@symbols);    # returns a binary string
+
+High-level function that generates an Adaptive Arithmetic Coding block.
+
+It takes two parameters: C<\@symbols>, which represents the symbols to be encoded, and C<$out_fh>, which is optional, and represents the file-handle where to write the result.
+
+When the second parameter is omitted, the function returns a binary string.
+
+=head2 decode_adaptive_ac_entry
+
+    my $symbols = decode_adaptive_ac_entry($fh);
+    my $symbols = decode_adaptive_ac_entry($string);
+
+Inverse of C<create_adaptive_ac_entry()>.
+
 =head2 lz77_compress
 
-    lz77_compress($data, $out_fh);       # writes to filehandle
+    lz77_compress($data, $out_fh);       # writes to file-handle
     my $string = lz77_compress($data);   # returns a binary string
 
 High-level function that performs LZ77 (Lempel-Ziv 1977) compression on the provided data, using the pipeline:
@@ -1850,11 +2227,11 @@ It takes a single parameter, C<$data>, representing the data string to be compre
 =head2 lzss_compress
 
     # With Huffman coding
-    lzss_compress($data, $out_fh);       # writes to filehandle
+    lzss_compress($data, $out_fh);       # writes to file-handle
     my $string = lzss_compress($data);   # returns a binary string
 
     # With Arithmetic Coding
-    lzss_compress($data, $out_fh, \&create_ac_entry);              # writes to filehandle
+    lzss_compress($data, $out_fh, \&create_ac_entry);              # writes to file-handle
     my $string = lzss_compress($data, undef, \&create_ac_entry);   # returns a binary string
 
 High-level function that performs LZSS (Lempel-Ziv-Storer-Szymanski) compression on the provided data, using the pipeline:
@@ -1866,11 +2243,11 @@ It takes a single parameter, C<$data>, representing the data string to be compre
 
 =head2 lz77_decompress / lzss_decompress
 
-    # Writing to filehandle
+    # Writing to file-handle
     lzss_decompress($fh, $out_fh);
     lzss_decompress($string, $out_fh);
 
-    # Writing to filehandle (does Arithmetic decoding)
+    # Writing to file-handle (does Arithmetic decoding)
     lzss_decompress($fh, $out_fh, \&decode_ac_entry);
     lzss_decompress($string, $out_fh, \&decode_ac_entry);
 
@@ -1886,7 +2263,7 @@ Inverse of C<lzss_compress()> and C<lz77_compress()>.
 
 =head2 lzw_compress
 
-    lzw_compress($data, $out_fh);       # writes to filehandle
+    lzw_compress($data, $out_fh);       # writes to file-handle
     my $string = lzw_compress($data);   # returns a binary string
 
 High-level function that performs LZW (Lempel-Ziv-Welch) compression on the provided data, using the pipeline:
@@ -1906,16 +2283,16 @@ It takes a single parameter, C<$data>, representing the data string to be compre
     my $data = lzw_decompress($fh);
     my $data = lzw_decompress($string);
 
-Performs Lempel-Ziv-Welch (LZW) decompression on the provided string or filehandle. Inverse of C<lzw_compress()>.
+Performs Lempel-Ziv-Welch (LZW) decompression on the provided string or file-handle. Inverse of C<lzw_compress()>.
 
 =head2 bz2_compress
 
     # Using Huffman Coding
-    bz2_compress($data, $out_fh);        # writes to filehandle
+    bz2_compress($data, $out_fh);        # writes to file-handle
     my $string = bz2_compress($data);    # returns a binary string
 
     # Using Arithmetic Coding
-    bz2_compress($data, $out_fh, \&create_ac_entry);               # writes to filehandle
+    bz2_compress($data, $out_fh, \&create_ac_entry);               # writes to file-handle
     my $string = bz2_compress($data, undef, \&create_ac_entry);    # returns a binary string
 
 High-level function that performs Bzip2-like compression on the provided data, using the pipeline:
@@ -1934,11 +2311,11 @@ When the additional optional argument, C<$out_fh>, is provided, the compressed d
 
 =head2 bz2_decompress
 
-    # Writes to filehandle
+    # Writes to file-handle
     bz2_decompress($fh, $out_fh);
     bz2_decompress($string, $out_fh);
 
-    # Writes to filehandle (does Arithmetic decoding)
+    # Writes to file-handle (does Arithmetic decoding)
     bz2_decompress($fh, $out_fh, \&decode_ac_entry);
     bz2_decompress($string, $out_fh, \&decode_ac_entry);
 
@@ -1955,11 +2332,11 @@ Inverse of C<bz2_compress()>.
 =head2 bz2_compress_symbolic
 
     # Does Huffman coding
-    bz2_compress_symbolic(\@symbols, $out_fh);      # writes to filehandle
+    bz2_compress_symbolic(\@symbols, $out_fh);      # writes to file-handle
     my $string = bz2_compress_symbolic(\@symbols);  # returns a binary string
 
     # Does Arithmetic coding
-    bz2_compress_symbolic(\@symbols, $out_fh, \&create_ac_entry);             # writes to filehandle
+    bz2_compress_symbolic(\@symbols, $out_fh, \&create_ac_entry);             # writes to file-handle
     my $string = bz2_compress_symbolic(\@symbols, undef, \&create_ac_entry);  # returns a binary string
 
 Similar to C<bz2_compress()>, except that it accepts an arbitrary array-ref of non-negative integer symbols as input. It is also a bit slower on large inputs.
@@ -1991,7 +2368,7 @@ It takes two parameters: C<\@integers>, representing the sequence of arbitrary i
 
 =head2 delta_decode
 
-    # Given a filehandle
+    # Given a file-handle
     my $integers = delta_decode($fh);
     my $integers = delta_decode($fh, 1);       # double
 
@@ -2009,7 +2386,7 @@ Encodes a sequence of non-negative integers using Fibonacci coding, returning a
 
 =head2 fibonacci_decode
 
-    # Given a filehandle
+    # Given a file-handle
     my $symbols = fibonacci_decode($fh);
 
     # Given a binary string
@@ -2017,6 +2394,22 @@ Encodes a sequence of non-negative integers using Fibonacci coding, returning a
 
 Inverse of C<fibonacci_encode()>.
 
+=head2 elias_gamma_encode
+
+    my $string = elias_gamma_encode(\@symbols);
+
+Encodes a sequence of non-negative integers using Elias Gamma coding, returning a binary string.
+
+=head2 elias_gamma_decode
+
+    # Given a file-handle
+    my $symbols = elias_gamma_decode($fh);
+
+    # Given a binary string
+    my $symbols = elias_gamma_decode($string);
+
+Inverse of C<elias_gamma_encode()>.
+
 =head2 elias_omega_encode
 
     my $string = elias_omega_encode(\@symbols);
@@ -2025,7 +2418,7 @@ Encodes a sequence of non-negative integers using Elias Omega coding, returning
 
 =head2 elias_omega_decode
 
-    # Given a filehandle
+    # Given a file-handle
     my $symbols = elias_omega_decode($fh);
 
     # Given a binary string
@@ -2044,12 +2437,12 @@ This method is particularly effective for encoding a sequence of integers that a
 =head2 abc_decode
 
     # Given a filehandle
-    my $nonnegative_integers = abc_decode($fh);
+    my $symbols = abc_decode($fh);
 
     # Given a binary string
-    my $nonnegative_integers = abc_decode($string);
+    my $symbols = abc_decode($string);
 
-Inverse of C<abc_decode()>.
+Inverse of C<abc_encode()>.
 
 =head2 bwt_encode
 
@@ -2060,7 +2453,7 @@ Applies the Burrows-Wheeler Transform (BWT) to a given string.
 
 It returns two values: C<$bwt>, which represents the transformed string, and C<$idx>, which holds the index of the original string in the sorted list of rotations.
 
-It takes an optional argument C<$lookahead_len>, which defaults to C<128>, representing the length of lookahead during sorting.
+It takes an optional argument C<$lookahead_len>, which defaults to C<128>, representing the length of look-ahead during sorting.
 
 =head2 bwt_decode
 
@@ -2106,26 +2499,42 @@ It takes two parameters: C<\@symbols>, representing the sequence of symbols to b
 
 Inverse of C<mtf_encode()>.
 
-=head2 mtf_encode_alphabet
+=head2 encode_alphabet
 
-    my $string = mtf_encode_alphabet(\@alphabet);
+    my $string = encode_alphabet(\@alphabet);
 
-Efficienlty encodes the MTF alphabet into a string.
+Efficienlty encodes an alphabet of symbols into a binary string.
 
-=head2 mtf_decode_alphabet
+=head2 decode_alphabet
 
-    my $alphabet = mtf_decode_alphabet($fh);
+    my $alphabet = decode_alphabet($fh);
+    my $alphabet = decode_alphabet($string);
 
-Decodes the MTF alphabet, given a filehandle C<$fh>, returning an array of symbols.
+Decodes an encoded alphabet, given a file-handle or a binary string, returning an array of symbols. Inverse of C<encode_alphabet()>.
+
+=head2 run_length
+
+    my $rl = run_length(\@symbols);
+    my $rl = run_length(\@symbols, $max_run);
+
+Performs Run-Length Encoding (RLE) on a sequence of symbolic elements.
+
+It takes two parameters: C<\@symbols>, representing an array of symbols, and C<$max_run>, indicating the maximum run length allowed.
+
+The function returns a 2D-array, with pairs: C<[symbol, run_length]>, such that the following code reconstructs the C<\@symbols> array:
+
+    my @symbols = map { ($_->[0]) x $_->[1] } @$rl;
+
+By default, the maximum run-length is unlimited.
 
 =head2 rle4_encode
 
-    my $rle4 = rle4_encode($symbols);
-    my $rle4 = rle4_encode($symbols, $max_run);
+    my $rle4 = rle4_encode(\@symbols);
+    my $rle4 = rle4_encode(\@symbols, $max_run);
 
 Performs Run-Length Encoding (RLE) on a sequence of symbolic elements, specifically designed for runs of four or more consecutive symbols.
 
-It takes two parameters: C<$symbols>, representing an array of symbols, and C<$max_run>, indicating the maximum run length allowed during encoding.
+It takes two parameters: C<\@symbols>, representing an array of symbols, and C<$max_run>, indicating the maximum run length allowed during encoding.
 
 The function returns the encoded RLE sequence as an array-ref of symbols.
 
@@ -2145,7 +2554,7 @@ Performs Zero-Run-Length Encoding (ZRLE) on a sequence of symbolic elements.
 
 It takes a single parameter C<$symbols>, representing the sequence of symbols to be encoded, and returns the encoded ZRLE sequence as an array-ref of symbols.
 
-This function efficiently encodes only runs of zeros, but also incremenets each symbol by C<1>.
+This function efficiently encodes only runs of zeros, but also increments each symbol by C<1>.
 
 =head2 zrle_decode
 
@@ -2159,19 +2568,42 @@ Inverse of C<zrle_encode()>.
 
 Performs Arithmetic Coding on the provided symbols.
 
-It takes a single parameter, C<\@symbols>, representing the symbols to be encoded. The function returns two values: C<$bitstring>, which is a string of 1s and 0s, and C<$freq>, representing the frequency table used for encoding.
+It takes a single parameter, C<\@symbols>, representing the symbols to be encoded.
+
+The function returns two values: C<$bitstring>, which is a string of 1s and 0s, and C<$freq>, representing the frequency table used for encoding.
 
 =head2 ac_decode
 
     my $symbols = ac_decode($bits_fh, \%freq);
     my $symbols = ac_decode($bitstring, \%freq);
 
-Performs Arithmetic Coding decoding using the provided frequency table and a string of 1s and 0s.
+Performs Arithmetic Coding decoding using the provided frequency table and a string of 1s and 0s. Inverse of C<ac_encode()>.
 
 It takes two parameters: C<$bitstring>, representing a string of 1s and 0s containing the arithmetic coded data, and C<\%freq>, representing the frequency table used for encoding.
 
 The function returns the decoded sequence of symbols.
 
+=head2 adaptive_ac_encode
+
+    my ($bitstring, $alphabet) = adaptive_ac_encode(\@symbols);
+
+Performs Adaptive Arithmetic Coding on the provided symbols.
+
+It takes a single parameter, C<\@symbols>, representing the symbols to be encoded.
+
+The function returns two values: C<$bitstring>, which is a string of 1s and 0s, and C<$alphabet>, which is an array-ref of distinct sorted symbols.
+
+=head2 adaptive_ac_decode
+
+    my $symbols = adaptive_ac_decode($bits_fh, \@alphabet);
+    my $symbols = adaptive_ac_decode($bitstring, \@alphabet);
+
+Performs Adaptive Arithmetic Coding decoding using the provided frequency table and a string of 1s and 0s.
+
+It takes two parameters: C<$bitstring>, representing a string of 1s and 0s containing the adaptive arithmetic coded data, and C<\@alphabet>, representing the array of distinct sorted symbols that appear in the encoded data.
+
+The function returns the decoded sequence of symbols.
+
 =head2 lzw_encode
 
     my $symbols = lzw_encode($string);
@@ -2194,15 +2626,27 @@ It takes a single parameter, C<\@symbols>, representing the encoded symbols to b
 
     my $bit = read_bit($fh, \$buffer);
 
-Reads a single bit from a filehandle C<$fh>.
+Reads a single bit from a file-handle C<$fh>.
 
 The function stores the extra bits inside the C<$buffer>, reading one character at a time from the filehandle.
 
 =head2 read_bits
 
-    my $bits = read_bits($fh, $bits_len);
+    my $bitstring = read_bits($fh, $bits_len);
 
-Reads a specified number of bits (C<$bits_len>) from a filehandle (C<$fh>) and returns them as a string.
+Reads a specified number of bits (C<$bits_len>) from a file-handle (C<$fh>) and returns them as a string.
+
+=head2 binary_vrl_encode
+
+    my $bitstring_enc = binary_vrl_encode($bitstring);
+
+Given a string of 1s and 0s, returns back a bitstring of 1s and 0s encoded using variable run-length encoding.
+
+=head2 binary_vrl_decode
+
+    my $bitstring = binary_vrl_decode($bitstring_enc);
+
+Given an encoded bitstring, returned by C<binary_vrl_encode()>, gives back the decoded string of 1s and 0s.
 
 =head2 bwt_sort
 
@@ -2211,7 +2655,7 @@ Reads a specified number of bits (C<$bits_len>) from a filehandle (C<$fh>) and r
 
 Low-level function that sorts the rotations of a given string using the Burrows-Wheeler Transform (BWT) algorithm.
 
-It takes two parameters: C<$string>, which is the input string to be transformed, and C<$LOOKAHEAD_LEN> (optional), representing the length of lookahead during sorting.
+It takes two parameters: C<$string>, which is the input string to be transformed, and C<$LOOKAHEAD_LEN> (optional), representing the length of look-ahead during sorting.
 
 The function returns an array-ref of indices.
 
@@ -2265,7 +2709,7 @@ Low-level function that performs LZSS (Lempel-Ziv-Storer-Szymanski) compression
 
 It takes a single parameter, C<$data>, representing the data string to be compressed.
 
-The function returns three values: C<$literals>, which is an array-ref of uncompressed bytes, C<$indices>, which contains the indices of the back-references, and C<$lengths>, which holds the lengths of the matched substrings.
+The function returns three values: C<$literals>, which is an array-ref of uncompressed bytes, C<$indices>, which contains the indices of the back-references, and C<$lengths>, which holds the lengths of the matched sub-strings.
 
 A back-reference is returned only when it's beneficial (i.e.: when it may not inflate the data). Otherwise, the corresponding index and length are both set to C<0>.
 
@@ -2279,7 +2723,7 @@ Low-level function that performs LZ77 (Lempel-Ziv 1977) compression on the provi
 
 It takes a single parameter, C<$data>, representing the data string to be compressed.
 
-The function returns three values: C<$literals>, which is an array-ref of uncompressed bytes, C<$indices>, which contains the indices of the matched substrings, and C<$lengths>, which holds the lengths of the matched substrings.
+The function returns three values: C<$literals>, which is an array-ref of uncompressed bytes, C<$indices>, which contains the indices of the matched sub-strings, and C<$lengths>, which holds the lengths of the matched sub-strings.
 
 Lengths are limited to C<255>.
 
@@ -2288,15 +2732,15 @@ Lengths are limited to C<255>.
     my $data = lz77_decode($literals, $indices, $lengths);
     my $data = lzss_decode($literals, $indices, $lengths);
 
-Low-level function that performs LZ77 (Lempel-Ziv 1977) decompression using the provided literals, indices, and lengths of matched substrings.
+Low-level function that performs LZ77 (Lempel-Ziv 1977) decompression using the provided literals, indices, and lengths of matched sub-strings.
 
-It takes three parameters: C<$literals>, representing the array-ref of uncompressed bytes, C<$indices>, containing the indices of the matched substrings, and C<$lengths>, holding the lengths of the matched substrings.
+It takes three parameters: C<$literals>, representing the array-ref of uncompressed bytes, C<$indices>, containing the indices of the matched sub-strings, and C<$lengths>, holding the lengths of the matched sub-strings.
 
 The function returns the decompressed data as a string.
 
 =head2 deflate_encode
 
-    # Writes to filehandle
+    # Writes to file-handle
     deflate_encode($size, \@literals, \@distances, \@lengths, $out_fh);
     deflate_encode($size, \@literals, \@distances, \@lengths, $out_fh, \&create_ac_entry);
 
